import random

APIKEY = "lstzFDEOhfFNMLikKa0am9mgEKLBl49T"

ASSET_TYPE = ["etf", "fund", "stock"]

EXCHANGE = {
    'ARCX' : 'NYSE Arca (Pacific Stock Exchange)',
    'BATS' : 'BATS Exchange inc',
    'CHIA' : 'Chi-X Australia',
    'E0WWE$$ALL' : 'WORLDWIDE_EQUITY',
    'FINR' : 'OTC Composite',
    'IPSX' : 'equity_exchange_IPSX',
    'IXUS' : 'Standard + Poor’s Indices',
    'MABX' : 'Madrid stock exchange Equities',
    'MSCO' : 'MSCI (EOD Only)',
    'MSTARFund' : 'Morningstar Fund',
    'OTCM' : 'OTC Markets (was Pinksheets)',
    'USCO' : 'US composite market',
    'XAMS' : 'NYSE Euronext Amsterdam Equities',
    'XASE' : 'American Stock Exchange',
    'XASX' : 'Australia Stock Exchange',
    'XATH' : 'Athens Stock Exchange',
    'XBER' : 'Berlin',
    'XBKK' : 'Thailand Stock Exchange',
    'XBOM' : 'Mumbai Stock Exchange',
    'XBRU' : 'NYSE Euronext Brussels Equities',
    'XCNQ' : 'CNSX (Canada National Stock Exchange)',
    'XCSE' : 'OMX exchange Copenhagen',
    'XDUB' : 'Irish Stock Exchange',
    'XDUS' : 'Düsseldorf',
    'XETR' : 'XETRA Frankfurt',
    'XEUR' : 'EUREX (was DTB Deutsche Termin Börse)',
    'XFRA' : 'Frankfurter Börse (or Germany in general)',
    'XHAM' : 'Hamburg',
    'XHAN' : 'Hannover (via CEF)',
    'XHEL' : 'Helsinki (Equities)',
    'XHKF' : 'Hong Kong Stock Exchange / Hong Kong Futures Exchange / Hang Seng Indices',
    'XHKG' : 'HONG_KONG',
    'XICE' : 'OMX Exchange Iceland',
    'XIST' : 'Istanbul Stock Exchange',
    'XKOS' : 'Korea Stock Exchange (KOSCOM)',
    'XLIS' : 'NYSE Euronext Lisbon',
    'XLIT' : 'OMX Exchange Vilnius',
    'XLON' : 'LSE (SEAQ national)',
    'XLUX' : 'Luxembourg Stock Exchange',
    'XMEX' : 'Bolsa de valores Mexico',
    'XMIL' : 'Milan Stock Exchange (Borsa Italia)',
    'XMUN' : 'Munchen',
    'XNAS' : 'NASDAQ national market',
    'XNSE' : 'India National Stock Exchange',
    'XNYS' : 'New York Stock Exchange',
    'XNZE' : 'New Zealand Stock Exchange',
    'XOSE' : 'Osaka',
    'XOSL' : 'Oslo bors',
    'XOTC' : 'OTCBB, over the counter bulletin board equities',
    'XPAR' : 'NYSE Euronext Paris Equities',
    'XRIS' : 'OMX Exchange Riga',
    'XSES' : 'singapore',
    'XSES' : 'Singapore Stock Exchange',
    'XSHE' : 'Shenzhen Stock Exchange',
    'XSHG' : 'Shanghai Stock Exchange',
    'XSTO' : 'NASDAQ OMX Stockholm',
    'XSTU' : 'Stuttgart',
    'XSWX' : 'SIX Feed (was SWX)',
    'XTAI' : 'Taiwan Stock Exchange',
    'XTAL' : 'Tallin Stock Exchange',
    'XTKS' : 'Tokyo Stock Exchange',
    'XTSE' : 'Toronto Stock Exchange',
    'XWAR' : 'Warsaw Stock Exchange',
    'XWBO' : 'Wiener Boerse',


}


FIELDS = [
    "AdministratorCompanyId",
    "AlphaM36",
    "AnalystRatingScale",
    "AverageCreditQualityCode",
    "AverageMarketCapital",
    "BetaM36",
    "BondStyleBox",
    "brandingCompanyId",
    "categoryId",
    "CategoryName",
    "ClosePrice",
    "currency",
    "DebtEquityRatio",
    "distribution",
    "DividendYield",
    "EBTMarginYear1",
    "EffectiveDuration",
    "EPSGrowth3YYear1",
    "equityStyle",
    "EquityStyleBox",
    "exchangeCode",
    "ExchangeId",
    "ExpertiseAdvanced",
    "ExpertiseBasic",
    "ExpertiseInformed",
    "FeeLevel",
    "fundShareClassId",
    "fundSize",
    "fundStyle",
    "FundTNAV",
    "GBRReturnD1",
    "GBRReturnM0",
    "GBRReturnM1",
    "GBRReturnM12",
    "GBRReturnM120",
    "GBRReturnM3",
    "GBRReturnM36",
    "GBRReturnM6",
    "GBRReturnM60",
    "GBRReturnW1",
    "geoRegion",
    "globalAssetClassId",
    "globalCategoryId",
    "iMASectorId",
    "IndustryName",
    "InitialPurchase",
    "instrumentName",
    "investment",
    "investmentExpertise",
    "investmentObjective",
    "investmentType",
    "investorType",
    "InvestorTypeEligibleCounterparty",
    "InvestorTypeProfessional",
    "InvestorTypeRetail",
    "LargestSector",
    "LegalName",
    "managementStyle",
    "ManagerTenure",
    "MarketCap",
    "MarketCountryName",
    "MaxDeferredLoad",
    "MaxFrontEndLoad",
    "MaximumExitCostAcquired",
    "MorningstarRiskM255",
    "Name",
    "NetMargin",
    "ongoingCharge",
    "OngoingCostActual",
    "PEGRatio",
    "PERatio",
    "PerformanceFeeActual",
    "PriceCurrency",
    "QuantitativeRating",
    "R2M36",
    "ReturnD1",
    "ReturnM0",
    "ReturnM1",
    "ReturnM12",
    "ReturnM120",
    "ReturnM3",
    "ReturnM36",
    "ReturnM6",
    "ReturnM60",
    "ReturnProfileGrowth",
    "ReturnProfileHedging",
    "ReturnProfileIncome",
    "ReturnProfileOther",
    "ReturnProfilePreservation",
    "ReturnW1",
    "RevenueGrowth3Y",
    "riskSrri",
    "ROATTM",
    "ROETTM",
    "ROEYear1",
    "ROICYear1",
    "SecId",
    "SectorName",
    "shareClassType",
    "SharpeM36",
    "StandardDeviationM36",
    "starRating",
    "StarRatingM255",
    "SustainabilityRank",
    "sustainabilityRating",
    "TenforeId",
    "Ticker",
    "totalReturn",
    "totalReturnTimeFrame",
    "TrackRecordExtension",
    "TransactionFeeActual",
    "umbrellaCompanyId",
    "Universe",
    "Yield_M12",
    "yieldPercent",
]


FILTER_FUND = [
    "AdministratorCompanyId",
    "AnalystRatingScale",
    "BondStyleBox",
    "BrandingCompanyId",
    "CategoryId",
    "CollectedSRRI",
    "distribution",
    "EquityStyleBox",
    "ExpertiseInformed",
    "FeeLevel",
    "FundTNAV",
    "GBRReturnM0",
    "GBRReturnM12",
    "GBRReturnM120",
    "GBRReturnM36",
    "GBRReturnM60",
    "GlobalAssetClassId",
    "GlobalCategoryId",
    "IMASectorID",
    "IndexFund",
    "InvestorTypeProfessional",
    "LargestRegion",
    "LargestSector",
    "OngoingCharge",
    "QuantitativeRating",
    "ReturnProfilePreservation",
    "ShareClassTypeId",
    "starRating",
    "SustainabilityRank",
    "UmbrellaCompanyId",
    "Yield_M12",
]

FILTER_STOCK = [
    "debtEquityRatio",
    "DividendYield",
    "epsGrowth3YYear1",
    "EquityStyleBox",
    "GBRReturnM0",
    "GBRReturnM12",
    "GBRReturnM36",
    "GBRReturnM60",
    "GBRReturnM120",
    "IndustryId",
    "MarketCap",
    "netMargin",
    "PBRatio",
    "PEGRatio",
    "PERatio",
    "PSRatio",
    "revenueGrowth3Y",
    "roattm",
    "roettm",
    "SectorId",
]


SITE = {
    "af": {"iso3": "AFG", "site": ""},
    "ax": {"iso3": "ALA", "site": ""},
    "al": {"iso3": "ALB", "site": ""},
    "dz": {"iso3": "DZA", "site": ""},
    "as": {"iso3": "ASM", "site": ""},
    "ad": {"iso3": "AND", "site": ""},
    "ao": {"iso3": "AGO", "site": ""},
    "ai": {"iso3": "AIA", "site": ""},
    "aq": {"iso3": "ATA", "site": ""},
    "ag": {"iso3": "ATG", "site": ""},
    "ar": {"iso3": "ARG", "site": ""},
    "am": {"iso3": "ARM", "site": ""},
    "aw": {"iso3": "ABW", "site": ""},
    "au": {"iso3": "AUS", "site": ""},
    "at": {"iso3": "AUT", "site": ""},
    "az": {"iso3": "AZE", "site": ""},
    "bs": {"iso3": "BHS", "site": ""},
    "bh": {"iso3": "BHR", "site": ""},
    "bd": {"iso3": "BGD", "site": ""},
    "bb": {"iso3": "BRB", "site": ""},
    "by": {"iso3": "BLR", "site": ""},
    "be": {"iso3": "BEL", "site": "https://www.morningstar.be/be/"},
    "bz": {"iso3": "BLZ", "site": ""},
    "bj": {"iso3": "BEN", "site": ""},
    "bm": {"iso3": "BMU", "site": ""},
    "bt": {"iso3": "BTN", "site": ""},
    "bo": {"iso3": "BOL", "site": ""},
    "bq": {"iso3": "BES", "site": ""},
    "ba": {"iso3": "BIH", "site": ""},
    "bw": {"iso3": "BWA", "site": ""},
    "bv": {"iso3": "BVT", "site": ""},
    "br": {"iso3": "BRA", "site": "https://www.morningstarbr.com/br/"},
    "io": {"iso3": "IOT", "site": ""},
    "bn": {"iso3": "BRN", "site": ""},
    "bg": {"iso3": "BGR", "site": ""},
    "bf": {"iso3": "BFA", "site": ""},
    "bi": {"iso3": "BDI", "site": ""},
    "cv": {"iso3": "CPV", "site": ""},
    "kh": {"iso3": "KHM", "site": ""},
    "cm": {"iso3": "CMR", "site": ""},
    "ca": {"iso3": "CAN", "site": "https://www.morningstar.ca/ca/"},
    "ky": {"iso3": "CYM", "site": ""},
    "cf": {"iso3": "CAF", "site": ""},
    "td": {"iso3": "TCD", "site": ""},
    "cl": {"iso3": "CHL", "site": "https://www.morningstar.cl/cl/"},
    "cn": {"iso3": "CHN", "site": ""},
    "cx": {"iso3": "CXR", "site": ""},
    "cc": {"iso3": "CCK", "site": ""},
    "co": {"iso3": "COL", "site": ""},
    "km": {"iso3": "COM", "site": ""},
    "cd": {"iso3": "COD", "site": ""},
    "cg": {"iso3": "COG", "site": ""},
    "ck": {"iso3": "COK", "site": ""},
    "cr": {"iso3": "CRI", "site": ""},
    "ci": {"iso3": "CIV", "site": ""},
    "hr": {"iso3": "HRV", "site": ""},
    "cu": {"iso3": "CUB", "site": ""},
    "cw": {"iso3": "CUW", "site": ""},
    "cy": {"iso3": "CYP", "site": ""},
    "cz": {"iso3": "CZE", "site": ""},
    "dk": {"iso3": "DNK", "site": "https://www.morningstar.dk/dk/"},
    "dj": {"iso3": "DJI", "site": ""},
    "dm": {"iso3": "DMA", "site": ""},
    "do": {"iso3": "DOM", "site": ""},
    "ec": {"iso3": "ECU", "site": ""},
    "eg": {"iso3": "EGY", "site": ""},
    "sv": {"iso3": "SLV", "site": ""},
    "gq": {"iso3": "GNQ", "site": ""},
    "er": {"iso3": "ERI", "site": ""},
    "ee": {"iso3": "EST", "site": ""},
    "sz": {"iso3": "SWZ", "site": ""},
    "et": {"iso3": "ETH", "site": ""},
    "fk": {"iso3": "FLK", "site": ""},
    "fo": {"iso3": "FRO", "site": ""},
    "fj": {"iso3": "FJI", "site": ""},
    "fi": {"iso3": "FIN", "site": "https://www.morningstar.fi/fi/"},
    "fr": {"iso3": "FRA", "site": "https://www.morningstar.fr/fr/"},
    "gf": {"iso3": "GUF", "site": ""},
    "pf": {"iso3": "PYF", "site": ""},
    "tf": {"iso3": "ATF", "site": ""},
    "ga": {"iso3": "GAB", "site": ""},
    "gm": {"iso3": "GMB", "site": ""},
    "ge": {"iso3": "GEO", "site": ""},
    "de": {"iso3": "DEU", "site": "https://www.morningstar.de/de/"},
    "gh": {"iso3": "GHA", "site": ""},
    "gi": {"iso3": "GIB", "site": ""},
    "gr": {"iso3": "GRC", "site": ""},
    "gl": {"iso3": "GRL", "site": ""},
    "gd": {"iso3": "GRD", "site": ""},
    "gp": {"iso3": "GLP", "site": ""},
    "gu": {"iso3": "GUM", "site": ""},
    "gt": {"iso3": "GTM", "site": ""},
    "gg": {"iso3": "GGY", "site": ""},
    "gn": {"iso3": "GIN", "site": ""},
    "gw": {"iso3": "GNB", "site": ""},
    "gy": {"iso3": "GUY", "site": ""},
    "ht": {"iso3": "HTI", "site": ""},
    "hm": {"iso3": "HMD", "site": ""},
    "va": {"iso3": "VAT", "site": ""},
    "hn": {"iso3": "HND", "site": ""},
    "hk": {"iso3": "HKG", "site": "https://www.morningstar.hk/hk/"},
    "hu": {"iso3": "HUN", "site": ""},
    "is": {"iso3": "ISL", "site": ""},
    "in": {"iso3": "IND", "site": ""},
    "id": {"iso3": "IDN", "site": ""},
    "ir": {"iso3": "IRN", "site": ""},
    "iq": {"iso3": "IRQ", "site": ""},
    "ie": {"iso3": "IRL", "site": "https://www.morningstarfunds.ie/ie/"},
    "im": {"iso3": "IMN", "site": ""},
    "il": {"iso3": "ISR", "site": ""},
    "it": {"iso3": "ITA", "site": "https://www.morningstar.it/it/"},
    "jm": {"iso3": "JAM", "site": ""},
    "jp": {"iso3": "JPN", "site": ""},
    "je": {"iso3": "JEY", "site": ""},
    "jo": {"iso3": "JOR", "site": ""},
    "kz": {"iso3": "KAZ", "site": ""},
    "ke": {"iso3": "KEN", "site": ""},
    "ki": {"iso3": "KIR", "site": ""},
    "kp": {"iso3": "PRK", "site": ""},
    "kr": {"iso3": "KOR", "site": ""},
    "kw": {"iso3": "KWT", "site": ""},
    "kg": {"iso3": "KGZ", "site": ""},
    "la": {"iso3": "LAO", "site": ""},
    "lv": {"iso3": "LVA", "site": ""},
    "lb": {"iso3": "LBN", "site": ""},
    "ls": {"iso3": "LSO", "site": ""},
    "lr": {"iso3": "LBR", "site": ""},
    "ly": {"iso3": "LBY", "site": ""},
    "li": {"iso3": "LIE", "site": ""},
    "lt": {"iso3": "LTU", "site": ""},
    "lu": {"iso3": "LUX", "site": ""},
    "mo": {"iso3": "MAC", "site": ""},
    "mk": {"iso3": "MKD", "site": ""},
    "mg": {"iso3": "MDG", "site": ""},
    "mw": {"iso3": "MWI", "site": ""},
    "my": {"iso3": "MYS", "site": ""},
    "mv": {"iso3": "MDV", "site": ""},
    "ml": {"iso3": "MLI", "site": ""},
    "mt": {"iso3": "MLT", "site": ""},
    "mh": {"iso3": "MHL", "site": ""},
    "mq": {"iso3": "MTQ", "site": ""},
    "mr": {"iso3": "MRT", "site": ""},
    "mu": {"iso3": "MUS", "site": ""},
    "yt": {"iso3": "MYT", "site": ""},
    "mx": {"iso3": "MEX", "site": "https://www.morningstar.com.mx/mx/"},
    "fm": {"iso3": "FSM", "site": ""},
    "md": {"iso3": "MDA", "site": ""},
    "mc": {"iso3": "MCO", "site": ""},
    "mn": {"iso3": "MNG", "site": ""},
    "me": {"iso3": "MNE", "site": ""},
    "ms": {"iso3": "MSR", "site": ""},
    "ma": {"iso3": "MAR", "site": ""},
    "mz": {"iso3": "MOZ", "site": ""},
    "mm": {"iso3": "MMR", "site": ""},
    "na": {"iso3": "NAM", "site": ""},
    "nr": {"iso3": "NRU", "site": ""},
    "np": {"iso3": "NPL", "site": ""},
    "nl": {"iso3": "NLD", "site": "https://www.morningstar.nl/nl/"},
    "nc": {"iso3": "NCL", "site": ""},
    "nz": {"iso3": "NZL", "site": ""},
    "ni": {"iso3": "NIC", "site": ""},
    "ne": {"iso3": "NER", "site": ""},
    "ng": {"iso3": "NGA", "site": ""},
    "nu": {"iso3": "NIU", "site": ""},
    "nf": {"iso3": "NFK", "site": ""},
    "mp": {"iso3": "MNP", "site": ""},
    "no": {"iso3": "NOR", "site": "https://www.morningstar.no/no/"},
    "om": {"iso3": "OMN", "site": ""},
    "pk": {"iso3": "PAK", "site": ""},
    "pw": {"iso3": "PLW", "site": ""},
    "ps": {"iso3": "PSE", "site": ""},
    "pa": {"iso3": "PAN", "site": ""},
    "pg": {"iso3": "PNG", "site": ""},
    "py": {"iso3": "PRY", "site": ""},
    "pe": {"iso3": "PER", "site": ""},
    "ph": {"iso3": "PHL", "site": ""},
    "pn": {"iso3": "PCN", "site": ""},
    "pl": {"iso3": "POL", "site": ""},
    "pt": {"iso3": "PRT", "site": "https://www.morningstar.pt/pt/"},
    "pr": {"iso3": "PRI", "site": ""},
    "qa": {"iso3": "QAT", "site": ""},
    "re": {"iso3": "REU", "site": ""},
    "ro": {"iso3": "ROU", "site": ""},
    "ru": {"iso3": "RUS", "site": ""},
    "rw": {"iso3": "RWA", "site": ""},
    "bl": {"iso3": "BLM", "site": ""},
    "sh": {"iso3": "SHN", "site": ""},
    "kn": {"iso3": "KNA", "site": ""},
    "lc": {"iso3": "LCA", "site": ""},
    "mf": {"iso3": "MAF", "site": ""},
    "pm": {"iso3": "SPM", "site": ""},
    "vc": {"iso3": "VCT", "site": ""},
    "ws": {"iso3": "WSM", "site": ""},
    "sm": {"iso3": "SMR", "site": ""},
    "st": {"iso3": "STP", "site": ""},
    "sa": {"iso3": "SAU", "site": ""},
    "sn": {"iso3": "SEN", "site": ""},
    "rs": {"iso3": "SRB", "site": ""},
    "sc": {"iso3": "SYC", "site": ""},
    "sl": {"iso3": "SLE", "site": ""},
    "sg": {"iso3": "SGP", "site": "https://sg.morningstar.com/sg/"},
    "sx": {"iso3": "SXM", "site": ""},
    "sk": {"iso3": "SVK", "site": ""},
    "si": {"iso3": "SVN", "site": ""},
    "sb": {"iso3": "SLB", "site": ""},
    "so": {"iso3": "SOM", "site": ""},
    "za": {"iso3": "ZAF", "site": ""},
    "gs": {"iso3": "SGS", "site": ""},
    "ss": {"iso3": "SSD", "site": ""},
    "es": {"iso3": "ESP", "site": "https://www.morningstar.es/es/"},
    "lk": {"iso3": "LKA", "site": ""},
    "sd": {"iso3": "SDN", "site": ""},
    "sr": {"iso3": "SUR", "site": ""},
    "sj": {"iso3": "SJM", "site": ""},
    "se": {"iso3": "SWE", "site": "https://www.morningstar.se/se/"},
    "ch": {"iso3": "CHE", "site": "https://www.morningstar.ch/ch/"},
    "sy": {"iso3": "SYR", "site": ""},
    "tw": {"iso3": "TWN", "site": ""},
    "tj": {"iso3": "TJK", "site": ""},
    "tz": {"iso3": "TZA", "site": ""},
    "th": {"iso3": "THA", "site": ""},
    "tl": {"iso3": "TLS", "site": ""},
    "tg": {"iso3": "TGO", "site": ""},
    "tk": {"iso3": "TKL", "site": ""},
    "to": {"iso3": "TON", "site": ""},
    "tt": {"iso3": "TTO", "site": ""},
    "tn": {"iso3": "TUN", "site": ""},
    "tr": {"iso3": "TUR", "site": ""},
    "tm": {"iso3": "TKM", "site": ""},
    "tc": {"iso3": "TCA", "site": ""},
    "tv": {"iso3": "TUV", "site": ""},
    "ug": {"iso3": "UGA", "site": ""},
    "ua": {"iso3": "UKR", "site": ""},
    "ae": {"iso3": "ARE", "site": ""},
    "gb": {"iso3": "GBR", "site": "https://www.morningstar.co.uk/uk/"},
    "um": {"iso3": "UMI", "site": ""},
    "us": {"iso3": "USA", "site": "https://www.morningstar.com/"},
    "uy": {"iso3": "URY", "site": ""},
    "uz": {"iso3": "UZB", "site": ""},
    "vu": {"iso3": "VUT", "site": ""},
    "ve": {"iso3": "VEN", "site": ""},
    "vn": {"iso3": "VNM", "site": ""},
    "vg": {"iso3": "VGB", "site": ""},
    "vi": {"iso3": "VIR", "site": ""},
    "wf": {"iso3": "WLF", "site": ""},
    "eh": {"iso3": "ESH", "site": ""},
    "ye": {"iso3": "YEM", "site": ""},
    "zm": {"iso3": "ZMB", "site": ""},
    "zw": {"iso3": "ZWE", "site": ""},
}


USER_AGENTS = [
    "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10.5; en-US; rv:1.9.1b3) Gecko/20090305 Firefox/3.1b3 GTB5",
    "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10.5; ko; rv:1.9.1b2) Gecko/20081201 Firefox/3.1b2",
    "Mozilla/5.0 (X11; U; SunOS sun4u; en-US; rv:1.9b5) Gecko/2008032620 Firefox/3.0b5",
    "Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.8.1.12) Gecko/20080214 Firefox/2.0.0.12",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; cs; rv:1.9.0.8) Gecko/2009032609 Firefox/3.0.8",
    "Mozilla/5.0 (X11; U; OpenBSD i386; en-US; rv:1.8.0.5) Gecko/20060819 Firefox/1.5.0.5",
    "Mozilla/5.0 (Windows; U; Windows NT 5.0; es-ES; rv:1.8.0.3) Gecko/20060426 Firefox/1.5.0.3",
    "Mozilla/5.0 (Windows; U; WinNT4.0; en-US; rv:1.7.9) Gecko/20050711 Firefox/1.0.6",
    "Mozilla/5.0 (Windows; Windows NT 6.1; rv:2.0b2) Gecko/20100720 Firefox/4.0b2",
    "Mozilla/5.0 (X11; Linux x86_64; rv:2.0b4) Gecko/20100818 Firefox/4.0b4",
    "Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.2) Gecko/20100308 Ubuntu/10.04 (lucid) Firefox/3.6 GTB7.1",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:2.0b7) Gecko/20101111 Firefox/4.0b7",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:2.0b8pre) Gecko/20101114 Firefox/4.0b8pre",
    "Mozilla/5.0 (X11; Linux x86_64; rv:2.0b9pre) Gecko/20110111 Firefox/4.0b9pre",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:2.0b9pre) Gecko/20101228 Firefox/4.0b9pre",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:2.2a1pre) Gecko/20110324 Firefox/4.2a1pre",
    "Mozilla/5.0 (X11; U; Linux amd64; rv:5.0) Gecko/20100101 Firefox/5.0 (Debian)",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:6.0a2) Gecko/20110613 Firefox/6.0a2",
    "Mozilla/5.0 (X11; Linux i686 on x86_64; rv:12.0) Gecko/20100101 Firefox/12.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:15.0) Gecko/20120716 Firefox/15.0a2",
    "Mozilla/5.0 (X11; Ubuntu; Linux armv7l; rv:17.0) Gecko/20100101 Firefox/17.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:21.0) Gecko/20130328 Firefox/21.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:22.0) Gecko/20130328 Firefox/22.0",
    "Mozilla/5.0 (Windows NT 5.1; rv:25.0) Gecko/20100101 Firefox/25.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.8; rv:25.0) Gecko/20100101 Firefox/25.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:28.0) Gecko/20100101 Firefox/28.0",
    "Mozilla/5.0 (X11; Linux i686; rv:30.0) Gecko/20100101 Firefox/30.0",
    "Mozilla/5.0 (Windows NT 5.1; rv:31.0) Gecko/20100101 Firefox/31.0",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:33.0) Gecko/20100101 Firefox/33.0",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:40.0) Gecko/20100101 Firefox/40.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.12; rv:58.0) Gecko/20100101 Firefox/58.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/525.19 (KHTML, like Gecko) Chrome/1.0.154.53 Safari/525.19",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/525.19 (KHTML, like Gecko) Chrome/1.0.154.36 Safari/525.19",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.10 (KHTML, like Gecko) Chrome/7.0.540.0 Safari/534.10",
    "Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/534.4 (KHTML, like Gecko) Chrome/6.0.481.0 Safari/534.4",
    "Mozilla/5.0 (Macintosh; U; Intel Mac OS X; en-US) AppleWebKit/533.4 (KHTML, like Gecko) Chrome/5.0.375.86 Safari/533.4",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/532.2 (KHTML, like Gecko) Chrome/4.0.223.3 Safari/532.2",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/532.0 (KHTML, like Gecko) Chrome/4.0.201.1 Safari/532.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/532.0 (KHTML, like Gecko) Chrome/3.0.195.27 Safari/532.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/530.5 (KHTML, like Gecko) Chrome/2.0.173.1 Safari/530.5",
    "Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/534.10 (KHTML, like Gecko) Chrome/8.0.558.0 Safari/534.10",
    "Mozilla/5.0 (X11; U; Linux x86_64; en-US) AppleWebKit/540.0 (KHTML,like Gecko) Chrome/9.1.0.0 Safari/540.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/534.14 (KHTML, like Gecko) Chrome/9.0.600.0 Safari/534.14",
    "Mozilla/5.0 (X11; U; Windows NT 6; en-US) AppleWebKit/534.12 (KHTML, like Gecko) Chrome/9.0.587.0 Safari/534.12",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.13 (KHTML, like Gecko) Chrome/9.0.597.0 Safari/534.13",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.16 (KHTML, like Gecko) Chrome/10.0.648.11 Safari/534.16",
    "Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US) AppleWebKit/534.20 (KHTML, like Gecko) Chrome/11.0.672.2 Safari/534.20",
    "Mozilla/5.0 (Windows NT 6.0) AppleWebKit/535.1 (KHTML, like Gecko) Chrome/14.0.792.0 Safari/535.1",
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/535.2 (KHTML, like Gecko) Chrome/15.0.872.0 Safari/535.2",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/535.7 (KHTML, like Gecko) Chrome/16.0.912.36 Safari/535.7",
    "Mozilla/5.0 (Windows NT 6.0; WOW64) AppleWebKit/535.11 (KHTML, like Gecko) Chrome/17.0.963.66 Safari/535.11",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_6_8) AppleWebKit/535.19 (KHTML, like Gecko) Chrome/18.0.1025.45 Safari/535.19",
    "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/535.24 (KHTML, like Gecko) Chrome/19.0.1055.1 Safari/535.24",
    "Mozilla/5.0 (Windows NT 6.2) AppleWebKit/536.6 (KHTML, like Gecko) Chrome/20.0.1090.0 Safari/536.6",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.1 (KHTML, like Gecko) Chrome/22.0.1207.1 Safari/537.1",
    "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.15 (KHTML, like Gecko) Chrome/24.0.1295.0 Safari/537.15",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.93 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/28.0.1467.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/30.0.1599.101 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/31.0.1623.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/34.0.1847.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/37.0.2062.103 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.38 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2490.71 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.62 Safari/537.36",
]


def random_user_agent():
    """
    This function selects a random User-Agent from the User-Agent list, . User-Agents are used in
    order to avoid the limitations of the requests to morningstar.com. The User-Agent is
    specified on the headers of the requests and is different for every request.



    Returns:
        :obj:`str` - user_agent:
            The returned :obj:`str` is the name of a random User-Agent, which will be passed on the
            headers of a request so to avoid restrictions due to the use of multiple requests from the
            same User-Agent.

    """

    return random.choice(USER_AGENTS)



# flake8: noqa

""" mstarpy init """
from .funds import Funds
from .stock import Stock
from .search import filter_universe, search_field, search_filter, search_funds, search_stock

__version__ = "2.0.0"



from .security import Security


class Stock(Security):
    """
    Main class to access data about stocks, inherit from Security class
    Args:
        term (str): text to find a stock, can be a name, part of a name or the isin of the stocks
        pageSize (int): number of stocks to return
        itemRange (int) : index of stocks to return (must be inferior to PageSize)
        proxies = (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

    Examples:
        >>> Stocks('0P0000712R', 9, 0)
        >>> Stocks('visa')

    Raises:
        TypeError: raised whenever the parameter type is not the type expected
        ValueError : raised whenever the parameter is not valid or no stock found

    """

    def __init__(
        self,
        term=None,
        exchange: str = "",
        pageSize: int = 1,
        itemRange: int = 0,
        filters: dict = {},
        proxies: dict = {},
    ):

        super().__init__(
            term=term,
            asset_type="stock",
            exchange=exchange,
            pageSize=pageSize,
            itemRange=itemRange,
            filters=filters,
            proxies=proxies,
        )

    def analysisData(self):
        """
        This function retrieves general data about the stock.

        Returns:
            dict with general data about stock

        Examples:
            >>> Stock("visa", exchange="nyse").analysisData()

        """
        return self.GetData("morningstarTake/v3", url_suffix="analysisData")

    def analysisReport(self):
        """
        This function retrieves the analysis of the stock.

        Returns:
            dict with analyst overview

        Examples:
            >>> Stock("visa", exchange="nyse").analysisReport()

        """
        return self.GetData("morningstarTake/v4", url_suffix="analysisReport")

    def balanceSheet(self, period="annual", reportType="original"):
        """
        This function retrieves the balance sheet.

        Args:
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated

        Returns:
            dict with balance sheet

        Examples:
            >>> Stock("visa", exchange="nyse").balanceSheet('quarterly', 'original')
            >>> Stock("visa", exchange="nyse").balanceSheet('annual', 'restated')

        """

        return self.financialStatement(
            "balancesheet", period=period, reportType=reportType
        )

    def boardOfDirectors(self):
        """
        This function retrieves information about the board of directors.

        Returns:
            dict with board of directors information

        Examples:
            >>> Stock("Alphabet Inc Class A").boardOfDirectors()

        """
        return self.GetData("insiders/boardOfDirectors")

    def cashFlow(self, period="annual", reportType="original"):
        """
        This function retrieves the cash flow.

        Args:
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated

        Returns:
            dict with cash flow

        Examples:
            >>> Stock("visa", exchange="nyse").cashFlow('annual', 'restated')
            >>> Stock("visa", exchange="nyse").cashFlow('quarterly', 'restated')

        """

        return self.financialStatement("cashflow", period=period, reportType=reportType)

    def dividends(self):
        """
        This function retrieves the dividends of the stock.

        Returns:
            dict with dividends

        Examples:
            >>> Stock("visa", exchange="nyse").dividends()

        """
        return self.GetData("dividends/v4")

    def esgRisk(self):
        """
        This function retrieves the esg risk of the stock.

        Returns:
            dict with esg risk

        Examples:
            >>> Stock("visa", exchange="nyse").esgRisk()

        """
        return self.GetData("esgRisk")

    def financialHealth(self):
        """
        This function retrieves the financial health of the stock.

        Returns:
            dict with financial health

        Examples:
            >>> Stock("visa", exchange="nyse").financialHealth()

        """
        return self.GetData("keyStats/financialHealth", url_suffix="")

    def financialStatement(
        self, statement="summary", period="annual", reportType="original"
    ):
        """
        This function retrieves the financial statement.

        Args:
            statement (str) : possible values are balancesheet, cashflow, incomestatement, summary
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated

        Returns:
            dict with financial statement

        Examples:
            >>> Stock("visa", exchange="nyse").financialStatement('summary', 'quarterly', 'original')
            >>> Stock("visa", exchange="nyse").financialStatement('cashflow', 'annual', 'restated')
            >>> Stock("visa", exchange="nyse").financialStatement('balancesheet', 'annual', 'restated')
            >>> Stock("visa", exchange="nyse").financialStatement('incomestatement', 'annual', 'restated')

        """

        if not isinstance(statement, str):
            raise TypeError("statement parameter should be a string")

        if not isinstance(period, str):
            raise TypeError("period parameter should be a string")

        if not isinstance(reportType, str):
            raise TypeError("reportType parameter should be a string")

        statement_choice = {
            "balancesheet": "balanceSheet",
            "cashflow": "cashFlow",
            "incomestatement": "incomeStatement",
            "summary": "summary",
        }

        period_choice = {"annual": "A", "quarterly": "Q"}

        reportType_choice = {"original": "A", "restated": "R"}

        if statement not in statement_choice:
            raise ValueError(
                f"statement parameter must take one of the following value : { ', '.join(statement_choice)}"
            )

        if period not in period_choice:
            raise ValueError(
                f"period parameter must take one of the following value : { ', '.join(period_choice)}"
            )

        if reportType not in reportType_choice:
            raise ValueError(
                f"reportType parameter must take one of the following value : { ', '.join(reportType_choice.keys())}"
            )

        params = {"reportType": reportType_choice[reportType]}
        if statement == "summary":
            return self.GetData(
                "newfinancials", params=params, url_suffix=f"{period}/summary"
            )

        params["dataType"] = period_choice[period]

        return self.GetData(
            "newfinancials",
            params=params,
            url_suffix=f"{statement_choice[statement]}/detail",
        )

    def financialSummary(self, period="annual", reportType="original"):
        """
        This function retrieves the financial statement summary.

        Args:
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated

        Returns:
            dict with financial statement summary

        Examples:
            >>> Stock("visa", exchange="nyse").financialSummary('quarterly', 'original')
            >>> Stock("visa", exchange="nyse").financialSummary('annual', 'restated')

        """

        return self.financialStatement("summary", period=period, reportType=reportType)

    def freeCashFlow(self):
        """
        This function retrieves the free cash flow.

        Returns:
            dict with free cash flow

        Examples:
            >>> Stock("visa", exchange="nyse").freeCashFlow()

        """
        return self.GetData("keyStats/cashFlow", url_suffix="")

    def historical(self, start_date, end_date, frequency="daily"):
        """
        This function retrieves the historical price, volume and divide of the stock.

        Args:
            start_date (datetime) : start date to get history
            end_date (datetime) : end date to get history
            frequency (str) : can be daily, weekly, monthly

        Returns:
            list of dict with price, volume and dividend

        Examples:
            >>> Stock("visa", exchange="nyse").history(datetime.datetime.today()- datetime.timedelta(30),datetime.datetime.today())

        """
        return self.TimeSeries(
            ["open", "high", "low", "close", "volume", "previousClose", "dividend"],
            start_date=start_date,
            end_date=end_date,
            frequency=frequency,
        )

    def incomeStatement(self, period="annual", reportType="original"):
        """
        This function retrieves the income statement.

        Args:
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated

        Returns:
            dict with income statement

        Examples:
            >>> Stock("visa", exchange="nyse").incomeStatement('quarterly', 'original')
            >>> Stock("visa", exchange="nyse").incomeStatement('annual', 'restated')

        """

        return self.financialStatement(
            "incomestatement", period=period, reportType=reportType
        )

    def institutionBuyers(self, top=20):
        """
        This function retrieves the institutions which buy the stock.

        Args:
            top (int) : number of institutions to return
        Returns:
            dict with the buyers

        Examples:
            >>> Stock("visa", exchange="nyse").institutionBuyers(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"Buyers/institution/{top}/data"
        )

    def institutionConcentratedOwners(self, top=20):
        """
        This function retrieves the institutions which are concentrated on the stock.

        Args:
            top (int) : number of institutions to return
        Returns:
            dict with the concentarted owners

        Examples:
            >>> Stock("visa", exchange="nyse").institutionConcentratedOwners(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"ConcentratedOwners/institution/{top}/data"
        )

    def institutionOwnership(self, top=20):
        """
        This function retrieves the main institutions which own the stock.

        Args:
            top (int) : number of institutions to return
        Returns:
            dict with the main owners

        Examples:
            >>> Stock("visa", exchange="nyse").institutionOwnership(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"OwnershipData/institution/{top}/data"
        )

    def institutionSellers(self, top=20):
        """
        This function retrieves the institutions which sell on the stock.

        Args:
            top (int) : number of institutions to return
        Returns:
            dict with sellers

        Examples:
            >>> Stock("visa", exchange="nyse").institutionSellers(top=50)

        """
        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"Sellers/institution/{top}/data"
        )

    def keyExecutives(self):
        """
        This function retrieves information oabout key excutives of the company.

        Returns:
            dict with key executives information

        Examples:
            >>> Stock("visa", exchange="nyse").keyExecutives()

        """
        return self.GetData("insiders/keyExecutives")

    def keyRatio(self):
        """
        This function retrieves the key ratio of the stock.

        Returns:
            dict with key ratio

        Examples:
            >>> Stock("visa", exchange="nyse").keyRatio()

        """
        return self.GetData("keyratios")

    def mutualFundBuyers(self, top=20):
        """
        This function retrieves the mutual funds which buy the stock.

        Args:
            top (int) : number of mutual funds to return
        Returns:
            dict with the buyers

        Examples:
            >>> Stock("visa", exchange="nyse").mutualFundBuyers(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData("ownership/v1", url_suffix=f"Buyers/mutualfund/{top}/data")

    def mutualFundConcentratedOwners(self, top=20):
        """
        This function retrieves the mutual funds which are concentrated on the stock.

        Args:
            top (int) : number of mutual funds to return
        Returns:
            dict with the concentarted owners

        Examples:
            >>> Stock("visa", exchange="nyse").mutualFundConcentratedOwners(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"ConcentratedOwners/mutualfund/{top}/data"
        )

    def mutualFundOwnership(self, top=20):
        """
        This function retrieves the main mutual funds which own the stock.

        Args:
            top (int) : number of mutual funds to return
        Returns:
            dict with the main owners

        Examples:
            >>> Stock("visa", exchange="nyse").mutualFundOwnership(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"OwnershipData/mutualfund/{top}/data"
        )

    def mutualFundSellers(self, top=20):
        """
        This function retrieves the mutual funds which sell on the stock.

        Args:
            top (int) : number of mutual funds to return
        Returns:
            dict with sellers

        Examples:
            >>> Stock("visa", exchange="nyse").mutualFundSellers(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"Sellers/mutualfund/{top}/data"
        )

    def operatingGrowth(self):
        """
        This function retrieves the operating growth of the stock.

        Returns:
            dict with operating growth

        Examples:
            >>> Stock("visa", exchange="nyse").operatingGrowth()

        """
        return self.GetData("keyStats/growthTable", url_suffix="")

    def operatingMargin(self):
        """
        This function retrieves the operating margin of the stock.

        Returns:
            dict with operating margin

        Examples:
            >>> Stock("visa", exchange="nyse").operatingMargin()

        """
        return self.GetData("keyStats/OperatingAndEfficiency", url_suffix="")

    def operatingPerformance(self):
        """
        This function retrieves the operating performance the stock.

        Returns:
            dict with voperating performance

        Examples:
            >>> Stock("visa", exchange="nyse").operatingPerformance()

        """
        return self.GetData("operatingPerformance/v2", url_suffix="")

    def split(self):
        """
        This function retrieves the split history of the stock.

        Returns:
            dict with split history

        Examples:
            >>> Stock("visa", exchange="nyse").split()

        """
        return self.GetData("split")

    def tradingInformation(self) -> dict:
        """
        This function retrieves the trading information of the stock
        such as Previous Close Price, Day Range, 52-Week Range, Bid/Ask,
        Market Cap, Volume/Avg
        Returns:
            dict with performance
        Examples:
            >>> Stock("visa", exchange="nyse").tradingInformation()

        """
        return self.RealtimeData("quotes")
    
    def trailingTotalReturn(self):
        """
        This function retrieves the performance of the stock and its index.

        Returns:
            dict with performance

        Examples:
            >>> Stock("visa", exchange="nyse").trailingTotalReturn()

        """
        return self.GetData("trailingTotalReturns")

    def transactionHistory(self):
        """
        This function retrieves the transaction of key people.

        Returns:
            list o dict of transaction of key people

        Examples:
            >>> Stock("visa", exchange="nyse").transactionHistory()

        """
        return self.GetData("insiders/transactionHistory")

    def transactionSummary(self):
        """
        This function retrieves the summuary of transactions of key people.

        Returns:
            list of dict with transactions

        Examples:
            >>> Stock("visa", exchange="nyse").transactionSummary()

        """
        return self.GetData("insiders/transactionChart")

    def valuation(self):
        """
        This function retrieves the valution of the stock.

        Returns:
            dict with valuation

        Examples:
            >>> Stock("visa", exchange="nyse").valuation()

        """
        return self.GetData("valuation", url_suffix="")



import requests
from bs4 import BeautifulSoup
import re
import warnings

from .utils import random_user_agent
from .utils import ASSET_TYPE, EXCHANGE, FIELDS, FILTER_FUND, FILTER_STOCK, SITE
from .error import not_200_response


def filter_universe(field=FILTER_FUND, proxies={}):
    """
    This function will use the screener of morningstar.co.uk 
    to find the possible filters and their values.

    Args:

      field (str | list) : field to find

    Returns:
      dict of filter
        {'LargestRegion': ['', 'RE_AfricaDeveloped', 'RE_AfricaEmerging', 
        'RE_Asia4TigersEmerging', 'RE_Asiaex4TigersEmerging', 'RE_Australasia', 
        'RE_Canada', 'RE_CentralandEasternEurope', 'RE_CentralandLatinAmericaEmerging', 
        'RE_Japan', 'RE_UnitedKingdom', 'RE_UnitedStates', 'RE_WesternEuropeEuro', 
        'RE_WesternEuropeNonEuroexUK'], 'SustainabilityRank': ['1', '2', '3', '4', '5']}

    Examples:
      >>> filter_universe(['LargestRegion','SustainabilityRank'])
      >>> filter_universe('FeeLevel')
      >>> filter_universe(FILTER_STOCK)

    """

    if not isinstance(field, (str, list)):
        raise TypeError("field parameter should be a string or a list")

    if "StarRatingM255" in field:
        raise ValueError("StarRatingM255 cannot be a field")

    if isinstance(field, list):
        filterDataPoints = "|".join(field)
    else:
        filterDataPoints = field

    if not isinstance(proxies, dict):
        raise TypeError("proxies parameter should be dict")

    params = {
        "outputType": "json",
        "filterDataPoints": filterDataPoints,
    }

    result = general_search(params, proxies=proxies)

    all_filter = {}

    if "filters" not in result:
        return all_filter
    if not result["filters"]:
        return all_filter

    for r in result["filters"][0]:
        all_filter[list(r)[0]] = r[list(r)[0]]

    return all_filter


def general_search(params, proxies={}):
    """
    This function will use the screener of morningstar.co.uk
    to find informations about funds or classification

    Args:
      params (dict) : paramaters of the request
      proxies (dict) : set the proxy if needed,
      example : {"http": "http://host:port","https": "https://host:port"}

    Returns:
      list of information

    Examples:
      >>> general_search(params = {
                                  'page' : 1,
                                  'pageSize' : 10,
                                  'sortOrder' : 'LegalName asc',
                                  'outputType' : 'json',
                                  'version' : 1,
                                  'universeIds' : 'FOFRA$$ALL',
                                  'currencyId': 'EUR',
                                  'securityDataPoints' : ['FundTNAV','GBRReturnD1'],
                                  'term' : 'myria',
                                  })


    """
    if not isinstance(params, dict):
        raise TypeError("params parameter should be dict")

    if not isinstance(proxies, dict):
        raise TypeError("proxies parameter should be dict")
    # url
    url = "https://tools.morningstar.co.uk/api/rest.svc/klr5zyak8x/security/screener"
    # headers
    headers = {
        "user-agent": random_user_agent(),
    }

    response = requests.get(url, params=params, headers=headers, proxies=proxies)

    not_200_response(url, response)

    return response.json()


def search_field(pattern=""):
    """
    This function retrieves the possible fields for the function dataPoint

    Args:
    pattern (str) : text contained in the field

    Returns:
        list of possible fields for the method dataPoint

    Example:
        >>> search_field('feE')
        >>> search_field('reTurn')

    """

    regex = re.compile(f"(?i){pattern}")
    filtered_list = list(filter(lambda field: regex.search(field), FIELDS))
    print(f"possible fields for function dataPoint can be : {', '.join(filtered_list)}")

    return filtered_list


def search_filter(pattern="", asset_type="fund"):
    """
    This function retrieves the possible filters for 
    the parameter filters of the function search_funds

    Args:
    pattern (str) : text contained in the filter

    Returns:
        list of possible filters

    Example:
        >>> search_filter('RetUrn')
        >>> search_filter('id')

    """
    if not isinstance(pattern, (str)):
        raise TypeError("pattern parameter should be a string")

    if not isinstance(asset_type, (str)):
        raise TypeError("asset_type parameter should be a string")

    if asset_type not in ASSET_TYPE:
        raise TypeError(
            f"asset_type parameter can only take one of the values : {','.join(ASSET_TYPE)}"
        )

    if asset_type == "stock":
        filter_type = FILTER_STOCK
    else:
        filter_type = FILTER_FUND

    regex = re.compile(f"(?i){pattern}")
    filtered_list = list(filter(lambda field: regex.search(field), filter_type))
    if asset_type == "stock":
        print(
            f"possible keys for the parameter filters of the method search_stock can be : {', '.join(filtered_list)}"
        )
    else:
        print(
            f"possible keys for the parameter filters of the method seach_funds can be : {', '.join(filtered_list)}"
        )

    return filtered_list


def search_funds(
    term, field, country="", pageSize=10, currency="EUR", filters={}, proxies={}
):
    """
    This function will use the screener of morningstar.co.uk
    to find funds which include the term.

    Args:
      term (str): text to find a funds can be a the name, 
      part of a name or the isin of the funds
      field (str | list) : field to find
      country (str) : text for code ISO 3166-1 alpha-2 of country
      pageSize (int): number of funds to return
      currency (str) : currency in 3 letters
      filters (dict) : filter funds, use the method filter_universe 
      to find the different possible filter keys and values
      proxies (dict) : set the proxy if needed , example : 
      {"http": "http://host:port","https": "https://host:port"}

    Returns:
      list of dict with fund information
        [{'SecId': 'F00000270E', 'TenforeId': '52.8.FR0010342600',
        'LegalName': '21 Gestion Active'}, {'SecId': 'F000013BGI',
        'TenforeId': '52.8.MT7000022612', 'LegalName':'24 Capital Management SICAV plc - 
        24 Global Currency Fund Share Class A USD Accumulation'}, 
        {'SecId': 'F00000PZHI', 'TenforeId': '52.8.FR0011443225', 
        'LegalName': '29 Haussmann Actions Europe C'}, {'SecId': 'F0GBR06QS1', 
        'TenforeId': '52.8.FR0007057427', 'LegalName': '29 Haussmann Actions Europe D'}, 
        {'SecId': 'F0000101BL', 'TenforeId': '52.8.FR0013266590', 'LegalName': 
        '29 Haussmann Actions Europe
        I'}, {'SecId': 'F00000JW7U', 'TenforeId': '52.8.LU0532306957', 
        'LegalName': '3F Generation Acc'}, {'SecId': 'F00000UDVR', 
        'TenforeId': '52.8.FR0011911189', 'LegalName': 'AAM Family Values E1'}, 
        {'SecId': 'F00000UDVS', 'TenforeId': '52.8.FR0011911197', 'LegalName': 'AAM Family Values
        I'}, {'SecId': 'F0GBR04RG5', 'TenforeId': '52.8.FR0007022025', 'LegalName': 
        'AAZ Capitalisation'}, 
        {'SecId': 'F000000ITD', 'TenforeId': '52.8.FR0010361600', 
        'LegalName': 'AAZ Prestige Or'}]

    Examples:
      >>> search_funds("Myria",['SecId','TenforeId','LegalName'],country="fr", pageSize=25)
      >>> search_funds("FR0011399914", 'LegalName', country="fr", pageSize=25)

    """

    if not isinstance(field, (str, list)):
        raise TypeError("field parameter should be a string or a list")

    if not isinstance(country, str):
        raise TypeError("country parameter should be a string")

    if country and country.lower() not in SITE:
        raise ValueError(
            f'country parameter can only take one of the values: {", ".join(SITE.keys())}'
        )

    if not isinstance(pageSize, int):
        raise TypeError("pageSize parameter should be an integer")

    if not isinstance(currency, str):
        raise TypeError("currency parameter should be a string")

    if not isinstance(filters, dict):
        raise TypeError("filters parameter should be a dict")

    if not isinstance(proxies, dict):
        raise TypeError("proxies parameter should be dict")

    if isinstance(field, list):
        securityDataPoints = "|".join(field)
    else:
        securityDataPoints = field
    # if country find iso
    if country:
        iso = SITE[country.lower()]["iso3"]
        universeIds = f"FO{iso}$$ALL"
    else:
        universeIds = ""

    filter_list = []
    # loop on filter dict
    for f in filters:
        if f not in FILTER_FUND:
            print(
                f"""{f} is not a valid filter and will be ignored. You can find the
                possible filters with the method search_filter()."""
            )
        else:
            # if list, IN condition
            if isinstance(filters[f], list):
                filter_list.append(f'{f}:IN:{":".join(filters[f])}')
            # if tuple, either, BTW, LT or GT condition
            elif isinstance(filters[f], tuple):
                if len(filters[f]) == 2:
                    if isinstance(filters[f][0], (int, float)):
                        filter_list.append(f"{f}:BTW:{filters[f][0]}:{filters[f][1]}")
                    elif filters[f][0] == "<":
                        filter_list.append(f"{f}:LT:{filters[f][1]}")
                    elif filters[f][0] == ">":
                        filter_list.append(f"{f}:GT:{filters[f][1]}")
            # else IN condition
            else:
                filter_list.append(f"{f}:IN:{filters[f]}")

    params = {
        "page": 1,
        "pageSize": pageSize,
        "sortOrder": "LegalName asc",
        "outputType": "json",
        "version": 1,
        "universeIds": universeIds,
        "currencyId": currency,
        "securityDataPoints": securityDataPoints,
        "term": term,
        "filters": "|".join(filter_list),
    }

    result = general_search(params, proxies=proxies)

    if result["rows"]:
        return result["rows"]
    else:
        print(f"0 fund found whith the term {term}")
        return {}


def search_stock(
    term, field, exchange = "E0WWE$$ALL", pageSize=10, currency="EUR", filters={}, proxies={}
):
    """
    This function will use the screener of morningstar.co.uk to find stocks which 
    include the term.

    Args:
      term (str): text to find a funds can be a the name, part of a name or the isin of the funds
      field (str | list) : field to find
      exchange (str) : stock echange closed list (.utils EXCHANGE)
      pageSize (int): number of funds to return
      currency (str) : currency in 3 letters
      filters (dict) : filter funds, use the method filter_universe 
      to find the different possible filter keys and values
      proxies (dict) : set the proxy if needed , example : 
      {"http": "http://host:port","https": "https://host:port"}

    Returns:
      list of dict with stocks information
        [{'SecId': '0P0001OMLZ', 'TenforeId': '126.1.VCXB', 'LegalName': 
        '10X Capital Venture Acquisition Corp III Ordinary Shares
        - Class A'}, {'SecId': '0P0001O9WE', 'TenforeId': '126.1.VCXB/U', 
        'LegalName': '10X Capital Venture Acquisition Corp III Units 
        (1 Ord Share Class A & 1/2 War)'}, {'SecId': '0P000184MI', 
        'TenforeId': '126.1.COE', 'LegalName': '51Talk无忧英语 ADR'}, 
        {'SecId': '0P0001NAQE', 'TenforeId': '126.1.AKA', 'LegalName': 
        'a.k.a. Brands Holding Corp'}]

    Examples:
      >>> search_stock("visa",['SecId','TenforeId','LegalName'],exchange="NYSE", pageSize=25)
      >>> search_stock("FR0000125486", 'LegalName', exchange="PARIS", pageSize=10)

    """

    if not isinstance(field, (str, list)):
        raise TypeError("field parameter should be a string or a list")

    if not isinstance(exchange, str):
        raise TypeError("exchange parameter should be a string")

    #by default, we look in all exchange
    universeIds = "E0WWE$$ALL"
    if exchange == "E0WWE$$ALL":
        pass
    #if we don't find the exchange
    elif not exchange.upper() in EXCHANGE:
        
        warnings.warn(f"""The exchange {exchange} is not found.
                      Exchange parameter can only take one of the following values : {", ".join(EXCHANGE)}.
                      The exchange was automaically set to E0WWE$$ALL to look in all exchanges.""")

    #determine universeIds according to exchange
    else:
        universeIds = f"E0EXG${exchange}"

    if not isinstance(pageSize, int):
        raise TypeError("pageSize parameter should be an integer")

    if not isinstance(currency, str):
        raise TypeError("currency parameter should be a string")

    if not isinstance(filters, dict):
        raise TypeError("filters parameter should be a dict")

    if not isinstance(proxies, dict):
        raise TypeError("proxies parameter should be dict")

    if isinstance(field, list):
        securityDataPoints = "|".join(field)
    else:
        securityDataPoints = field

    


    filter_list = []
    # loop on filter dict
    for f in filters:
        if f not in FILTER_STOCK:
            print(
                f"""{f} is not a valid filter and will be ignored.
                You can find the possible filters with the method search_filter()."""
            )
        else:
            # if list, IN condition
            if isinstance(filters[f], list):
                filter_list.append(f'{f}:IN:{":".join(filters[f])}')
            # if tuple, either, BTW, LT or GT condition
            elif isinstance(filters[f], tuple):
                if len(filters[f]) == 2:
                    if isinstance(filters[f][0], (int, float)):
                        filter_list.append(f"{f}:BTW:{filters[f][0]}:{filters[f][1]}")
                    elif filters[f][0] == "<":
                        filter_list.append(f"{f}:LT:{filters[f][1]}")
                    elif filters[f][0] == ">":
                        filter_list.append(f"{f}:GT:{filters[f][1]}")
            # else IN condition
            else:
                filter_list.append(f"{f}:IN:{filters[f]}")

    params = {
        "page": 1,
        "pageSize": pageSize,
        "sortOrder": "LegalName asc",
        "outputType": "json",
        "version": 1,
        "universeIds": universeIds,
        "currencyId": currency,
        "securityDataPoints": securityDataPoints,
        "term": term,
        "filters": "|".join(filter_list),
    }

    result = general_search(params, proxies=proxies)

    if result["rows"]:
        return result["rows"]
    else:
        print(f"0 stock found whith the term {term}")
        return {}


def token_chart(proxies={}):
    """
    This function will scrape the Bearer Token needed to access MS API chart data

    Args:
    proxies (dict) : set the proxy if needed ,
    example : {"http": "http://host:port","https": "https://host:port"}

    Returns:
    str bearer token

    """

    if not isinstance(proxies, dict):
        raise TypeError("proxies parameter should be dict")

    url = "https://www.morningstar.com/funds/xnas/afozx/chart"

    headers = {"user-agent": random_user_agent()}

    response = requests.get(url, headers=headers, proxies=proxies)

    all_text = response.text
    if all_text.find("token") == -1:
        return None

    token_start = all_text[all_text.find("token") :]
    return token_start[7 : token_start.find("}") - 1]


def token_fund_information(proxies={}):
    """
    This function will scrape the Bearer Token needed to access MS API funds information

    Args:
    proxies (dict) : set the proxy if needed , example :
    {"http": "http://host:port","https": "https://host:port"}

    Returns:
    str bearer token

    """
    if not isinstance(proxies, dict):
        raise TypeError("proxies parameter should be dict")

    url = "https://www.morningstar.co.uk/Common/funds/snapshot/PortfolioSAL.aspx"

    headers = {"user-agent": random_user_agent()}

    response = requests.get(url, headers=headers, proxies=proxies, timeout=120)
    soup = BeautifulSoup(response.text, "html.parser")
    script = soup.find_all("script", {"type": "text/javascript"})
    bearerToken = (
        str(script).rsplit("tokenMaaS:",maxsplit=1)[-1].split("}")[0].replace('"', "").strip()
    )
    return bearerToken


def token_investment_strategy(proxies={}):
    """
    This function will scrape the Bearer Token needed to access the investment strategy

    Args:
    proxies (dict) : set the proxy if needed , example :
    {"http": "http://host:port","https": "https://host:port"}

    Returns:
    str bearer token

    """
    if not isinstance(proxies, dict):
        raise TypeError("proxies parameter should be dict")

    url = "https://www.morningstar.com.au/investments/security/ASX/VHY"

    headers = {"user-agent": random_user_agent()}

    response = requests.get(url, headers=headers, proxies=proxies, timeout=120)

    all_text = response.text
    if all_text.find("token") == -1:
        return None
    start_flag = ',"'
    end_flag = '"www.morningstar.com.au"'
    token_end = all_text[: all_text.find(end_flag) - 3]
    return token_end[token_end.rfind(start_flag) + 2 :]



import datetime
import re
import requests

from .error import not_200_response
from .search import search_funds, search_stock, token_chart
from .utils import APIKEY, SITE, random_user_agent


# with Universe field, we can detect the asset class
# done find all stock echange and create a search_equity method
# add parameter exchange to Security class and search stocks if stock and exchange


class Security:
    """
    Parent class to access data about security

    Args:
        term (str): text to find a fund can be a name, part of a name or the isin of the funds
        country (str) : text for code ISO 3166-1 alpha-2 of country, should be '' for etf
        pageSize (int): number of funds to return
        itemRange (int) : index of funds to return (must be inferior to PageSize)
        proxies = (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

    Examples:
        >>> Security('0P0000712R', "ca", 9, 0)
        >>> Security('visa', "", 25, 2)

    Raises:
        TypeError: raised whenever the parameter type is not the type expected
        ValueError : raised whenever the parameter is not valid or no fund found

    """

    def __init__(
        self,
        term=None,
        asset_type: str = "",
        country: str = "",
        exchange: str = "",
        pageSize: int = 1,
        itemRange: int = 0,
        filters: dict = {},
        proxies: dict = {},
    ):

        if not isinstance(country, str):
            raise TypeError("country parameter should be a string")

        if country and not country.lower() in SITE.keys():
            raise ValueError(
                f'country parameter can only take one of the values: {", ".join(SITE.keys())}'
            )

        if not isinstance(pageSize, int):
            raise TypeError("pageSize parameter should be an integer")

        if not isinstance(itemRange, int):
            raise TypeError("itemRange parameter should be an integer")

        if pageSize <= itemRange:
            raise ValueError(
                "itemRange parameter should be strictly inferior to pageSize parameter"
            )

        if not isinstance(filters, dict):
            raise TypeError("filters parameter should be dict")

        if not isinstance(proxies, dict):
            raise TypeError("proxies parameter should be dict")

        self.proxies = proxies

        if country:
            self.site = SITE[country.lower()]["site"]
        else:
            self.site = ""

        self.country = country

        self.exchange = exchange

        self.asset_type = "security"

        code_list = []

        if exchange:
            code_list = search_stock(
                term,
                ["fundShareClassId", "SecId", "TenforeId", "LegalName", "Universe"],
                exchange=exchange,
                pageSize=pageSize,
                filters=filters,
                proxies=self.proxies,
            )
        else:
            code_list = search_funds(
                term,
                ["fundShareClassId", "SecId", "TenforeId", "LegalName", "Universe"],
                country,
                pageSize,
                filters=filters,
                proxies=self.proxies,
            )

        if code_list:
            if itemRange < len(code_list):
                self.code = code_list[itemRange]["fundShareClassId"]
                self.name = code_list[itemRange]["LegalName"]
                if "TenforeId" in code_list[itemRange]:
                    tenforeId = code_list[itemRange]["TenforeId"]
                    regex = re.compile("[0-9]*\.[0-9]\.")
                    self.isin = regex.sub("", tenforeId)
                else:
                    self.isin = None

                universe = code_list[itemRange]["Universe"]

                if universe[:2] == "E0":
                    self.asset_type = "stock"
                elif universe[:2] == "ET":
                    self.asset_type = "etf"
                elif universe[:2] == "FO":
                    self.asset_type = "fund"

                if universe[:2] == "E0" and asset_type in ["etf", "fund"]:
                    raise ValueError(
                        f"The security found with the term {term} is a stock and the parameter asset_type is equal to {asset_type}, the class Stock should be used with this security."
                    )

                if universe[:2] in ["FO", "ET"] and asset_type == "stock":
                    if universe[:2] == "FO":
                        raise ValueError(
                            f"The security found with the term {term} is a fund and the parameter asset_type is equal to {asset_type}, the class Fund should be used with this security."
                        )
                    else:
                        raise ValueError(
                            f"The security found with the term {term} is an ETF and the parameter asset_type is equal to {asset_type}, the class Fund should be used with this security."
                        )

            else:
                raise ValueError(
                    f"Found only {len(code_list)} {self.asset_type} with the term {term}. The paramater itemRange must maximum equal to {len(code_list)-1}"
                )
        else:
            if country:
                raise ValueError(
                    f"0 {self.asset_type} found with the term {term} and country {country}"
                )
            elif exchange:
                raise ValueError(
                    f"0 {self.asset_type} found with the term {term} and exchange {exchange}"
                )
            else:
                raise ValueError(f"0 {self.asset_type} found with the term {term}")

    def GetData(self, field, params={}, headers={}, url_suffix="data"):
        """
        Generic function to use MorningStar global api.

        Args:
            field (str) : endpoint of the request
            params (dict) : parameter for the request
            headers (dict) : headers of the request
            url_suffix (str) : suffixe of the url

        Raises:
            TypeError raised whenever type of paramater are invalid

        Returns:
            dict with data

        Examples:
            >>> Security("rmagx", "us").GetData("price/feeLevel")

        """

        if not isinstance(field, str):
            raise TypeError("field parameter should be a string")

        if not isinstance(params, dict):
            raise TypeError("params parameter should be a dict")

        if not isinstance(url_suffix, str):
            raise TypeError("url_suffix parameter should be a string")

        # url of API
        url = f"""https://api-global.morningstar.com/sal-service/v1/{self.asset_type}/{field}/{self.code}"""

        if url_suffix:
            url += f"""/{url_suffix}"""

        # headers
        default_headers = {
            "apikey": APIKEY,
        }

        all_headers = default_headers | headers

        response = requests.get(
            url, params=params, headers=all_headers, proxies=self.proxies
        )

        not_200_response(url, response)

        return response.json()

    def ltData(self, field, currency="EUR"):
        """
        Generic function to use MorningStar lt api.

        Args:
            field (str) : viewId in the params
            currency (str) : currency in 3 letters

        Raises:
            TypeError raised whenever type of paramater are invalid

        Returns:
            dict with data

        Examples:
            >>> Security("rmagx", "us").ltData("price/feeLevel")

        """
        if not isinstance(field, str):
            raise TypeError("field parameter should be a string")

        # url of API
        url = f"""https://lt.morningstar.com/api/rest.svc/klr5zyak8x/security_details/{self.code}"""

        params = {
            "viewId": field,
            "currencyId": currency,
            "itype": "msid",
            "languageId": "en",
            "responseViewFormat": "json",
        }
        response = requests.get(url, params=params, proxies=self.proxies)

        not_200_response(url, response)

        # responseis a list
        response_list = response.json()
        if response_list:
            return response_list[0]
        else:
            return {}

    def RealtimeData(self, url_suffix: str) -> dict:
        """
        This function retrieves historical data of the specified fields

        Args:
            url_suffix (str) : suffixe of the url

        Returns:
            dict of realtime data

        Examples:
            >>> Stock("visa", "us").RealtimeData("quotes")

        Raises:
            TypeError: raised whenever the parameter type 
            is not the type expected
            ConnectionError : raised whenever the response is not 200 OK

        """
        # error raised if url_suffix is not a string
        if not isinstance(url_suffix, str):
            raise TypeError("url_suffix parameter should be a string or a list")
        # url for realtime data
        url = f"""https://www.morningstar.com/api/v2/stores/realtime/{url_suffix}?securities={self.code}"""

        # header with user agent
        headers = {
                    'user-agent': random_user_agent(), 
                    }
        # response
        response = requests.get(url, headers=headers, proxies=self.proxies,
                                timeout=60)
        # manage response
        not_200_response(url, response)
        # result
        return response.json()
    
    def TimeSeries(self, field, start_date, end_date, frequency="daily"):
        """
        This function retrieves historical data of the specified fields

        Args:
            start_date (datetime) : start date to get history
            end_date (datetime) : end date to get history
            frequency (str) : can be daily, weekly, monthly

        Returns:
            list of dict time series

        Examples:
            >>> Funds("RMAGX", "us").TimeSeries(["nav","totalReturn"],datetime.datetime.today()- datetime.timedelta(30),datetime.datetime.today())

        Raises:
            TypeError: raised whenever the parameter type is not the type expected
            ValueError : raised whenever the parameter is not valid or no funds found

        """

        # error raised if field is not a string or a list
        if not isinstance(field, (str, list)):
            raise TypeError("field parameter should be a string or a list")

        # error raised if start_date is note a datetime.date
        if not isinstance(start_date, datetime.date):
            raise TypeError("start_date parameter should be a datetime.date")

        # error raised if end_date is note a datetime.date
        if not isinstance(end_date, datetime.date):
            raise TypeError("end_date parameter should be a datetime.date")

        # error if end_date < start_date
        if end_date < start_date:
            raise ValueError("end_date must be more recent than start_date")

        # error raised if frequency is not a string
        if not isinstance(frequency, str):
            raise TypeError("frequency parameter should be a string")

        # dict of frequency
        frequency_row = {"daily": "d", "weekly": "w", "monthly": "m"}

        # raise an error if frequency is not daily, wekly or monthly
        if frequency not in frequency_row:
            raise ValueError(
                f"frequency parameter must take one of the following value : { ', '.join(frequency_row.keys())}"
            )

        if isinstance(field, list):
            queryField = ",".join(field)
        else:
            queryField = field

        # bearer token
        bearer_token = token_chart()
        # url for nav
        url = f"https://www.us-api.morningstar.com/QS-markets/chartservice/v2/timeseries?query={self.code}:{queryField}&frequency={frequency_row[frequency]}&startDate={start_date.strftime('%Y-%m-%d')}&endDate={end_date.strftime('%Y-%m-%d')}&trackMarketData=3.6.3&instid=DOTCOM"
        # header with bearer token
        headers = {
            "user-agent": random_user_agent(),
            "authorization": f"Bearer {bearer_token}",
        }
        # response
        response = requests.get(url, headers=headers, proxies=self.proxies)
        # manage response
        not_200_response(url, response)
        # result
        result = response.json()
        # return empty list if we don't get data
        if not result:
            return []
        if "series" in result[0]:
            return result[0]["series"]

        return []



""" class funds """

import re
from bs4 import BeautifulSoup
import pandas as pd

import requests

from .error import no_site_error, not_200_response
from .search import search_funds, token_investment_strategy
from .utils import random_user_agent

from .security import Security


class Funds(Security):
    """
    Main class to access data about funds and etf, inherit from Security class

    Args:
        term (str): text to find a fund, can be a name, part of a name or the isin of the funds
        country (str) : text for code ISO 3166-1 alpha-2 of country, should be '' for etf
        pageSize (int): number of funds to return
        itemRange (int) : index of funds to return (must be inferior to PageSize)
        proxies = (dict) : set the proxy if needed ,
        example : {"http": "http://host:port","https": "https://host:port"}

    Examples:
        >>> Funds('0P0000712R', "ca", 9, 0)
        >>> Funds('bond', "uk", 25, 2)

    Raises:
        TypeError: raised whenever the parameter type is not the type expected
        ValueError : raised whenever the parameter is not valid or no fund found

    """

    def __init__(
        self,
        term=None,
        country: str = "",
        pageSize: int = 1,
        itemRange: int = 0,
        filters: dict = {},
        proxies: dict = {},
    ):

        super().__init__(
            term=term,
            asset_type="fund",
            country=country,
            pageSize=pageSize,
            itemRange=itemRange,
            filters=filters,
            proxies=proxies,
        )

    def allocationMap(self):
        """
        This function retrieves the asset allocation of the funds, index and category.

        Returns:
            dict with allocation map

        Examples:
            >>> Funds("myria", "fr").allocationMap()

        """
        return self.GetData("process/asset/v2")

    def allocationWeighting(self):
        """
        This function retrieves the Growth/Blend/Value and
        market capitalizaton allocation size of the funds.

        Returns:
            dict with allocation

        Examples:
            >>> Funds("myria", "fr").allocationWeighting()

        """
        return self.GetData("process/weighting")

    def analystRating(self):
        """
        This function retrieves the rating of the funds

        Returns:
            list of dict with ratings

            >>> Funds("RMAGX", "us").analystRating()

        """

        return self.GetData("parent/analystRating")

    def analystRatingTopFunds(self):
        """
        This function retrieves the rating Top funds

        Returns:
            dict with ratings

            >>> Funds("RMAGX", "us").analystRatingTopFunds()

        """

        return self.GetData("parent/analystRating/topfunds")

    def analystRatingTopFundsUpDown(self):
        """
        This function retrieves the rating funds Up Down

        Returns:
            dict with ratings

            >>> Funds("RMAGX", "us").analystRatingTopFundsUpDown()

        """

        return self.GetData("parent/analystRating/topfundsUpDown")



    def AnnualPerformance(self, cat):
        """
        This function retrieves the annual performance of the funds,
        index, category or the annual rank of the funds.

        Args:
            cat (str) : possible values are category, funds, index, rank
        Returns:
            dict annual performance or rank

        Raises:
            ValueError : raised whenever parameter cat is not category, funds, index, or rank

        Examples:
            >>> Funds("myria", "fr").AnnualPerformance("category")
            >>> Funds("myria", "fr").AnnualPerformance("funds")
            >>> Funds("myria", "fr").AnnualPerformance("index")
            >>> Funds("myria", "fr").AnnualPerformance("rank")

        """
        no_site_error(self.code, self.name, self.country, self.site)

        cat_row = {"funds": 0, "category": 1, "index": 2, "rank": 3}
        if cat not in cat_row:
            raise ValueError(
                f"cat parameter must take one of the following value : { ', '.join(cat_row.keys())}"
            )

        result = {}
        # headers random agent
        headers = {"user-agent": random_user_agent()}
        # page 1 - performance
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=1"

        response = requests.get(url, headers=headers, proxies=self.proxies)
        not_200_response(url, response)
        soup = BeautifulSoup(response.text, "html.parser")
        # label are dates
        regex = re.compile(".*heading number")
        label_list = soup.find(id="returnsCalenderYearDiv").find_all(
            "td", {"class": regex}
        )
        # funds performance, category performance, index performance, rank in category
        regex = re.compile(".*value number")
        # values
        value_list = soup.find(id="returnsCalenderYearDiv").find_all(
            "td", {"class": regex}
        )

        regex = re.compile("-|\/")
        # first col is nothing
        for i in range(1, len(label_list)):
            label = label_list[i].text
            # if today

            if regex.search(label):
                label = "current"
            # add category to label
            if label:

                result[f"{cat}_annual_performance_{label}"] = value_list[
                    i + (cat_row[cat]) * (len(label_list) - 1) - 1
                ].text

        return result

    def benchmark(self):
        """
        This function retrieves the benchmark name of the funds.

        Returns:
            str benchmark name

        Examples:
            >>> Funds("myria", "fr").benchmark()

        """
        return self.referenceIndex("benchmark")

    def carbonMetrics(self):
        """
        This function retrieves the carbon metrics of the funds.

        Returns:
            dict with carbon metrics

        Examples:
            >>> Funds("myria", "fr").carbonMetrics()

        """

        return self.GetData("esg/carbonMetrics")

    def category(self):
        """
        This function retrieves the category name of the funds.

        Returns:
            str category name

        Examples:
            >>> Funds("myria", "fr").category()

        """
        return self.referenceIndex("category")

    def categoryAnnualPerformance(self):
        """
        This function retrieves the annual performance of the category.

        Returns:
            dict annual performance of the category


        Examples:
            >>> Funds("myria", "fr").categoryAnnualPerformance()

        """

        return self.AnnualPerformance("category")

    def categoryCumulativePerformance(self):
        """
        This function retrieves the cumulative performance of the category.

        Returns:
            dict cumulative performance of the category

        Examples:
            >>> Funds("myria", "fr").categoryCumulativePerformance()

        """

        return self.CumulativePerformance("category")

    def contact(self):
        """
        This function retrieves information about the asset manager.

        Returns:
            dict contact

        Examples:
            >>> Funds("myria", "fr").contact()

        """
        no_site_error(self.code, self.name, self.country, self.site)
        result = {}
        # headers random agent
        headers = {"user-agent": random_user_agent()}
        # page 1 - performance
        # page 4 - info about found
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=4"
        response = requests.get(url, headers=headers, proxies=self.proxies)
        not_200_response(url, response)
        soup = BeautifulSoup(response.text, "html.parser")
        # label
        label_list = soup.find(id="managementManagementDiv").find_all(
            "td", {"class": "col1 label"}
        )
        # value
        value_list = soup.find(id="managementManagementDiv").find_all(
            "td", {"class": "col2 value number"}
        )
        for i in range(0, len(value_list)):
            label = label_list[i].text

            result[label] = value_list[i].text

        return result

    def costIllustration(self):
        """
        This function retrieves the cost of the funds.

        Returns:
            dict cost of funds

        Examples:
            >>> Funds("FOUSA00E5P", "us").costIllustration()

        """
        return self.GetData("price/costIllustration")

    def couponRange(self):
        """
        This function retrieves the coupon of the funds, index and category.

        Returns:
            dict coupon

        Examples:
            >>> Funds("rmagx", "us").couponRange()

        """
        return self.GetData("process/couponRange")

    def creditQuality(self):
        """
        This function retrieves the credit notation of the funds, index and category.

        Returns:
            dict credit notation

        Examples:
            >>> Funds("rmagx", "us").creditQuality()

        """
        return self.GetData("portfolio/creditQuality")

    def dataPoint(self, field, currency="EUR"):
        """
        This function retrieves infos about funds such as name,
        performance, risk metrics...

        Args:
        field (str or list) : field to find
        currency (str) : currency in 3 letters

        Returns:
            list of dict funds infos

        Example:
            >>> Funds("myria", "fr").dataPoint(['largestSector', 'Name', 'ongoingCharge'])
            >>> Funds("myria", "fr").dataPoint('SharpeM36')

        """
        return search_funds(
            self.code, field, self.country, 10, currency, proxies=self.proxies
        )

    def distribution(self, period="annual"):
        """
        This function retrieves the coupon distributed by the funds.

        Args:
            period (str) : annual or latest

        Raises:
            ValueError whenever the pariod parameter is not annual or latest

        Returns:
            dict distribution of coupon

        Example:
            >>> Funds("rmagx", "us").distribution("annual")

        """

        period_choice = ["annual", "latest"]
        if period not in period_choice:
            raise ValueError(
                f"""period parameter can only take one of
                             the values: {", ".join(period_choice)}"""
            )

        return self.GetData(f"distribution/{period}")

    def CumulativePerformance(self, cat):
        """
        This function retrieves the cumulative performance of funds, index and category.

        Args:
            cat (str) : possible values are category, funds, index

        Returns:
            dict cumulative performance

        Examples:
            >>> Funds("myria", "fr").CumulativePerformance("funds")
            >>> Funds("myria", "fr").CumulativePerformance("index")
            >>> Funds("myria", "fr").CumulativePerformance("category")

        """
        no_site_error(self.code, self.name, self.country, self.site)
        cat_row = {"funds": 2, "category": 3, "index": 4}
        if cat not in cat_row:
            raise ValueError(
                f"""cat parameter must take
                             one of the following value : { ', '.join(cat_row.keys())}"""
            )
        result = {}

        # headers random agent
        headers = {"user-agent": random_user_agent()}
        # page 1 - performance
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=1"

        response = requests.get(url, headers=headers, proxies=self.proxies)
        not_200_response(url, response)
        soup = BeautifulSoup(response.text, "html.parser")
        cumulative_performance_date = (
            soup.find(id="returnsTrailingDiv")
            .find("td", {"class": "titleBarNote"})
            .text
        )
        result["cumulative_performance_date"] = cumulative_performance_date
        # days
        regex = re.compile(".*label")
        label_list = soup.find(id="returnsTrailingDiv").find_all("td", {"class": regex})

        # cumulative performance cat
        regex = re.compile(f".*col{str(cat_row[cat])} value number")
        value_list = soup.find(id="returnsTrailingDiv").find_all("td", {"class": regex})
        # loop on label
        for i in range(0, len(label_list)):
            # label
            label = label_list[i].text
            # perf funds
            result[f"{cat}_cumulative_performance_{label}"] = re.sub(
                "[^0-9,-\.]", "", value_list[i].text
            )

        return result

    def equityStyle(self):
        """
        This function retrieves the equity style of the funds and category.

        Returns:
            dict equity style

        Examples:
            >>> Funds("myria", "fr").equityStyle()

        """
        return self.GetData("process/stockStyle/v2")

    def equityStyleBoxHistory(self):
        """
        This function retrieves the equity style history of the funds

        Returns:
            dict equity style history

        Examples:
            >>> Funds("myria", "fr").equityStyleBoxHistory()

        """
        return self.GetData("process/equityStyleBoxHistory")

    def esgData(self):
        """
        This function retrieves ESG data of the funds and category

        Returns:
            dict ESG data

        Examples:
            >>> Funds("myria", "fr").esgData()

        """

        return self.GetData("esg/v1")

    def factorProfile(self):
        """
        This function retrieves the factor profile of the funds, index and category

        Returns:
            dict factor profile

        Examples:
            >>> Funds("myria", "fr").factorProfile()

        """
        return self.GetData("factorProfile")

    def feeLevel(self):
        """
        This function retrieves the fees of the fund compare to its category.

        Returns:
            dict fees

        Examples:
            >>> Funds("rmagx", "us").feeLevel()

        """
        return self.GetData("price/feeLevel")

    def feeMifid(self, currency="EUR"):
        """
        This function retrieves the fees of the fund.

        Returns:
            dict fees

        Examples:
            >>> Funds("myria", "fr").feeMifid()

        """
        return self.ltData("Mifid", currency=currency)

    def fees(self):
        """
        This function retrieves the fees of the fund (by scraping pages);

        Returns:
            dict fees

        Examples:
            >>> Funds("myria", "fr").fees()

        """
        no_site_error(self.code, self.name, self.country, self.site)
        result = {}
        # headers random agent
        headers = {"user-agent": random_user_agent()}
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=5"
        response = requests.get(url, headers=headers, proxies=self.proxies)
        not_200_response(url, response)
        soup = BeautifulSoup(response.text, "html.parser")
        if soup.find(id="managementFeesDiv") == None:
            return {}
        # label
        label_list = soup.find(id="managementFeesDiv").find_all(
            "td", {"class": "label"}
        )
        # value
        value_list = soup.find(id="managementFeesDiv").find_all(
            "td", {"class": "value number"}
        ) + soup.find(id="managementFeesDiv").find_all(
            "td", {"class": "value number jdpa"}
        )
        for i in range(0, len(value_list)):
            label = label_list[i].text
            result[label] = re.sub("(\\n +)|(\\n)", "", value_list[i].text)

        return result

    def financialMetrics(self):
        """
        This function retrieves the final metrics of the funds and category.

        Returns:
            dict financial metrics

        Examples:
            >>> Funds("rmagx", "us").financialMetrics()

        """
        return self.GetData("process/financialMetrics")

    def fixedIncomeStyle(self):
        """
        This function retrieves the fixed income style of the funds and category.

        Returns:
            dict fixed income style

        Examples:
            >>> Funds("rmagx", "us").fixedIncomeStyle()

        """

        return self.GetData("process/fixedIncomeStyle")

    def fixedincomeStyleBoxHistory(self):
        """
        This function retrieves the fixed income style history of the funds.

        Returns:
            dict fixed income style

        Examples:
            >>> Funds("rmagx", "us").fixedincomeStyleBoxHistory()

        """
        return self.GetData("process/fixedincomeStyleBoxHistory")

    def fundsAnnualPerformance(self):
        """
        This function retrieves the annual performance of the funds.

        Returns:
            dict funds annual performance

        Examples:
            >>> Funds("myria", "fr").fundsAnnualPerformance()

        """
        return self.AnnualPerformance("funds")

    def fundsAnnualRank(self):
        """
        This function retrieves the annual rank of the funds in percentile.

        Returns:
            dict funds annual rank

        Examples:
            >>> Funds("myria", "fr").fundsAnnualRank()

        """
        return self.AnnualPerformance("rank")

    def fundsCumulativePerformance(self):
        """
        This function retrieves the cumulative performance of the funds.

        Returns:
            dict funds cumulative performance

        Examples:
            >>> Funds("myria", "fr").fundsCumulativePerformance()

        """
        return self.CumulativePerformance("funds")

    def fundsQuarterlyPerformance(self):
        """
        This function retrieves the quarterly performance of the funds.

        Returns:
            dict funds quarterly performance

        Examples:
            >>> Funds("myria", "fr").fundsCumulativePerformance()

        """
        no_site_error(self.code, self.name, self.country, self.site)
        result = {}
        # headers random agent
        headers = {"user-agent": random_user_agent()}
        # page 1 - performance
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=1"

        response = requests.get(url, headers=headers, proxies=self.proxies)
        not_200_response(url, response)
        soup = BeautifulSoup(response.text, "html.parser")
        quarterly_performance_date = (
            soup.find(id="returnsTrailingDiv")
            .find("td", {"class": "titleBarNote"})
            .text
        )
        result["quarterly_performance_date"] = quarterly_performance_date

        # quarter label
        regex = re.compile(".*heading number")
        quarter_list = soup.find(id="returnsQuarterlyDiv").find_all(
            "td", {"class": regex}
        )
        # year label
        regex = re.compile(".*label")
        year_list = soup.find(id="returnsQuarterlyDiv").find_all("td", {"class": regex})
        # 1st Quarter
        regex = re.compile(".*col2 value number")
        quarter_1_list = soup.find(id="returnsQuarterlyDiv").find_all(
            "td", {"class": regex}
        )
        # 2nd Quarter
        regex = re.compile(".*col3 value number")
        quarter_2_list = soup.find(id="returnsQuarterlyDiv").find_all(
            "td", {"class": regex}
        )
        # 3rd Quarter
        regex = re.compile(".*col4 value number")
        quarter_3_list = soup.find(id="returnsQuarterlyDiv").find_all(
            "td", {"class": regex}
        )
        # 4th Quarter
        regex = re.compile(".*col5 value number")
        quarter_4_list = soup.find(id="returnsQuarterlyDiv").find_all(
            "td", {"class": regex}
        )
        # loop on year
        for i in range(0, len(year_list)):
            label = "performance_%s_" % (year_list[i].text)
            result[label + "quarter_1"] = quarter_1_list[i].text
            result[label + "quarter_2"] = quarter_2_list[i].text
            result[label + "quarter_3"] = quarter_3_list[i].text
            result[label + "quarter_4"] = quarter_4_list[i].text
        return result

    def graphData(self):
        """
        This function retrieves historical data of the funds.

        Returns:
            dict historical data

        Examples:
            >>> Funds("myria", "fr").graphData()

        """

        return self.GetData("parent/graphData")

    def historicalData(self):
        """
        This function retrieves the historical price of the funds, index and category

        Returns:
            dict with historical data

        Examples:
            >>> Funds("myria", "fr").historicalData()

        """
        return self.GetData("performance/v3", url_suffix="")

    def historicalExpenses(self):
        """
        This function retrieves historical expenses of the funds.

        Returns:
            dict historical expenses

        Examples:
            >>> Funds("rmagx", "us").historicalExpenses()

        """
        if self.asset_type == "etf":
            return {}
        return self.GetData("price/historicalExpenses")

    def holdings(self, holdingType: str = "all"):
        """
        This function retrieves holdings of the funds.

        Args:
            holdingType (str) : paramater to select the kind of holdings; all, bond, equity or other

        Returns:
            pandas DataFrame holdings

        Raises:
            ValueError whenever the parameter is not all, bond, equity or other

        Examples:
            >>> Funds("myria", "fr").holdings("all")
            >>> Funds("myria", "fr").holdings("bond")
            >>> Funds("myria", "fr").holdings("equity")
            >>> Funds("myria", "fr").holdings("other")

        """
        holdingType_to_holdingPage = {
            "all": "all",
            "bond": "boldHoldingPage",
            "equity": "equityHoldingPage",
            "other": "otherHoldingPage",
        }
        if holdingType not in holdingType_to_holdingPage:
            raise ValueError(
                f"""parameter holdingType must take one of the following value
                : {", ".join(holdingType_to_holdingPage.keys())}"""
            )

        if holdingType == "all":
            return pd.DataFrame(
                self.position()["equityHoldingPage"]["holdingList"]
                + self.position()["boldHoldingPage"]["holdingList"]
                + self.position()["otherHoldingPage"]["holdingList"]
            )
        else:
            return pd.DataFrame(
                self.position()[holdingType_to_holdingPage[holdingType]]["holdingList"]
            )

    def indexAnnualPerformance(self):
        """
        This function retrieves the annual performance of the index.

        Returns:
            dict annual performance of the index


        Examples:
            >>> Funds("myria", "fr").indexAnnualPerformance()

        """
        return self.AnnualPerformance("index")

    def indexCumulativePerformance(self):
        """
        This function retrieves the cumulative performance of the index.

        Returns:
            dict cumulative performance of the index


        Examples:
            >>> Funds("myria", "fr").indexCumulativePerformance()

        """

        return self.CumulativePerformance("index")

    def investmentStrategy(self):
        """
        This function retrieves the investment strategy.

        Returns:
            dict investment strategy

        Examples:
            >>> Funds("LU0823421689").investmentStrategy()

        """
        bearer_token = token_investment_strategy()
        # header with bearer token
        headers = {
            "authorization": f"Bearer {bearer_token}",
        }
        return self.GetData("morningstarTake/investmentStrategy", headers=headers)

    def investmentLookup(self, currency="EUR"):
        """
        This function gives details about fund investment.

        Returns:
            dict fund investment

        Examples:
            >>> Funds("myria", "fr").investmentLookup()

        """
        return self.ltData("investmentTypeLookup", currency=currency)

    def keyStats(self):
        """
        This function retrieves the key status information of the funds,
        index, category or the annual rank of the funds.

        Returns:
            dict annual performance or rank

        Raises:
            ValueError : raised whenever parameter cat is not category,
            funds, index, or rank

        Examples:
            >>> Funds("myria", "fr").keyStats()

        """
        # headers random agent
        headers = {"user-agent": random_user_agent()}
        # page 1 - performance
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}"

        response = requests.get(url, headers=headers, proxies=self.proxies)
        not_200_response(url, response)
        soup = BeautifulSoup(response.text, "html.parser")

        table_rows = soup.find(id="overviewQuickstatsDiv").find_all("tr")
        details = []
        for row in table_rows:
            key, value, date_stamp = None, None, None
            for cell in row.find_all("td"):
                if cell.has_attr("class"):
                    if "line" in cell["class"]:
                        if "heading" in cell["class"]:
                            key = cell.get_text(separator="\n", strip=True)
                            if "\n" in key:
                                key, date_stamp = key.split("\n")
                        if "text" in cell["class"]:
                            value = cell.text
            if key and value and date_stamp:
                details.append({key: value, "date": date_stamp})
            elif key and value:
                details.append({key: value})

        return details
    def marketCapitalization(self):
        """
        This function retrieves the marketCapitalization breakdown of the funds,
        category and index.

        Returns:
            dict market capitalization

        Examples:
            >>> Funds("myria", "fr").marketCapitalization()

        """
        return self.GetData("process/marketCap")

    def maturitySchedule(self):
        """
        This function retrieves the maturity breakdown of the funds and category.

        Returns:
            dict maturity

        Examples:
            >>> Funds("rmagx", "us").maturitySchedule()

        """
        return self.GetData("process/maturitySchedule")

    def maxDrawDown(self, year=3):
        """
        This function retrieves the max drawdown of the funds, index and category.

        Args:
            year (int) : period of calculation in year

        Returns:
            dict max drawdown

        Raises:
            TypeError whenever the parameter year is not an integer

        Examples:
            >>> Funds("myria", "fr").maxDrawDown()

        """

        if not isinstance(year, int):
            raise TypeError("year parameter should be an integer")

        return self.GetData(
            "performance/marketVolatilityMeasure", params={"year": year}
        )

    def morningstarAnalyst(self):
        """
        This function retrieves the raiting of MorningStar analyst.

        Returns:
            dict rating

        Examples:
            >>> Funds("rmagx", "us").morningstarAnalyst()

        """

        return self.GetData("morningstarAnalyst")

    def multiLevelFixedIncomeData(
        self, primary="superEffectiveDuration", secondary="superSector.weight"
    ):
        """
        This function retrieves the exposures of fixed income

        Args:
            primary (str) : can be superEffectiveDuration, superYieldToWorst, creditQuality
            secondary (str) : can be superSector.weight, region.weight, creditQuality.weight

        Returns:
            dict exposures

        Raises:
            ValueError whenever the primary and seconday paramater is not a value expected

        Examples:
            >>> Funds("rmagx", "us").multiLevelFixedIncomeData()

        """

        primary_choice = [
            "superEffectiveDuration",
            "superYieldToWorst",
            "creditQuality",
        ]
        secondary_choice = [
            "superSector.weight",
            "region.weight",
            "creditQuality.weight",
        ]

        if primary not in primary_choice:
            raise ValueError(
                f"""primary parameter can only take one of the 
                values : {", ".join(primary_choice)}"""
            )

        if secondary not in secondary_choice:
            raise ValueError(
                f"""secondary parameter can only take one of the values 
                : {", ".join(secondary_choice)}"""
            )

        if primary == "creditQuality" and secondary == "creditQuality.weight":
            raise ValueError(
                f"primary and secondary parameters cannot be both credit quality"
            )

        return self.GetData(
            "multiLevelFixedIncomeData",
            params={"primary": primary, "secondary": secondary},
        )

    def nav(self, start_date, end_date, frequency="daily"):
        """
        This function retrieves the NAV of the funds

        Args:
            start_date (datetime) : start date to get nav
            end_date (datetime) : end date to get nav
            frequency (str) : can be daily, weekly, monthly

        Returns:
            list of dict with nav

            >>> Funds("RMAGX", "us").nav(datetime.datetime.today() 
            - datetime.timedelta(30),datetime.datetime.today())

        Raises:
            TypeError: raised whenever the parameter type is not the type expected
            ValueError : raised whenever the parameter is not valid or no funds found

        """

        return self.TimeSeries(
            ["nav", "totalReturn"],
            start_date=start_date,
            end_date=end_date,
            frequency=frequency,
        )

    def objectiveInvestment(self):
        """
        This function retrieves the objective of investment of the fund (by scraping pages);

        Returns:
            str objective investment

        Examples:
            >>> Funds("myria", "fr").objectiveInvestment()

        """

        no_site_error(self.code, self.name, self.country, self.site)
        # headers random agent
        headers = {"user-agent": random_user_agent()}
        # Page 1 - overview
        # url page overview
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}"
        # get HTML page overview
        response = requests.get(url, headers=headers, proxies=self.proxies)
        # if page not found

        not_200_response(url, response)
        # html page as soup
        soup = BeautifulSoup(response.text, "html.parser")
        # investment objective funds
        return (
            soup.find(id="overviewObjectiveDiv")
            .find("td", {"class": "value text"})
            .text
        )

    def otherFee(self):
        """
        This function retrieves the other fee of the etf

        Returns:
            dict fees

        Examples:
            >>> Funds("American Century Foc Dynmc Gr ETF").otherFee()

            {'expenseWaiver': False, 'expenseReimbursement': None, 
            'expirationDate': None, 'expenseWaivers': None}

        """
        return self.GetData("price/otherFee")

    def ownershipZone(self):
        """
        This function retrieves ownershipZone of the funds, index and category.

        Returns:
            dict ownershipZone

        Examples:
            >>> Funds("myria", "fr").ownershipZone()

        """

        return self.GetData("process/ownershipZone")

    def parentMstarRating(self):
        """
        This function retrieves the raiting of parent by MorningStar analyst.

        Returns:
            list of dict rating

        Examples:
            >>> Funds("rmagx", "us").parentMstarRating()

        """

        return self.GetData("parent/parentMstarRating")

    def parentSummary(self):
        """
        This function retrieves info about the parent.

        Returns:
            dict parent info

        Examples:
            >>> Funds("rmagx", "us").parentSummary()

        """
        return self.GetData("parent/parentSummary")

    def people(self):
        """
        This function retrieves info about people who works in the company.

        Returns:
            dict people info

        Examples:
            >>> Funds("rmagx", "us").people()

        """
        return self.GetData("people")

    def position(self):
        """
        This function retrieves the hodings of the funds.

        Returns:
            dict holdings

        Examples:
            >>> Funds("myria", "fr").position()

        """

        return self.GetData(
            "portfolio/holding/v2", params={"premiumNum": 10000, "freeNum": 10000}
        )

    def proxyVotingManagement(self):
        """
        This function retrieves the vote of management.

        Returns:
            dict vote

        Examples:
            >>> Funds("rmagx", "us").proxyVotingManagement()

        """
        return self.GetData("people/proxyVoting/management")

    def proxyVotingShareHolder(self):
        """
        This function retrieves the vote of shareholders.

        Returns:
            dict vote

        Examples:
            >>> Funds("rmagx", "us").proxyVotingShareHolder()

        """
        return self.GetData("people/proxyVoting/shareHolder")

    def productInvolvement(self):
        """
        This function retrieves the involvement of the funds

        Returns:
            dict involvement

        Examples:
            >>> Funds("myria", "fr").proxyVotingShareHolder()

        """

        return self.GetData("esg/productInvolvement")

    def referenceIndex(self, index):
        """
        This function retrieves the name of the category or the benchmark

        Args:
            index (str) : possible values are benchmark, category

        Returns:
            str category or benchmark

        Raises:
            ValueErrror whenever the index parameter is not category of benchmark

        Examples:
            >>> Funds("myria", "fr").referenceIndex("category")
            >>> Funds("myria", "fr").referenceIndex("benchmark")

        """
        no_site_error(self.code, self.name, self.country, self.site)

        index_row = {"benchmark": 0, "category": 1}
        if index not in index_row:
            raise ValueError(
                f"""index parameter must take one of the following value
                : { ', '.join(index_row.keys())}"""
            )

            # headers random agent
        headers = {"user-agent": random_user_agent()}
        # Page 1 - overview
        # url page overview
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}"
        # get HTML page overview
        response = requests.get(url, headers=headers, proxies=self.proxies)
        # if page not found

        not_200_response(url, response)

        # html page as soup
        soup = BeautifulSoup(response.text, "html.parser")
        benchmark_soup = soup.find(id="overviewBenchmarkDiv2Cols").find_all(
            "td", {"class": "value text"}
        )
        return benchmark_soup[index_row[index]].text

    def regionalSector(self):
        """
        This function retrieves the breakdown of the funds, category and index by region

        Returns:
            dict regional breakdown

        Examples:
            >>> Funds("myria", "fr").regionalSector()

        """
        return self.GetData("portfolio/regionalSector")

    def regionalSectorIncludeCountries(self):
        """
        This function retrieves the breakdown of the funds,
        category and index by region and country

        Returns:
            dict regional breakdown

        Examples:
            >>> Funds("myria", "fr").regionalSectorIncludeCountries()

        """
        return self.GetData("portfolio/regionalSectorIncludeCountries")

    def riskReturnScatterplot(self):
        """
        This function retrieves the return and standard
        deviation of the funds and category

        Returns:
            dict risk return

        Examples:
            >>> Funds("rmagx", "us").riskReturnScatterplot()

        """
        return self.GetData("performance/riskReturnScatterplot")

    def riskReturnSummary(self):
        """
        This function retrieves the return and risk summary 
        of the funds compare to the category

        Returns:
            dict risk return

        Examples:
            >>> Funds("rmagx", "us").riskReturnSummary()

        """

        return self.GetData("performance/riskReturnSummary")

    def riskVolatility(self):
        """
        This function retrieves the alpha, beta, R², 
        volatility and Sharpe ratio of the funds, category and index.

        Returns:
            dict econometrics

        Examples:
            >>> Funds("rmagx", "us").riskVolatility()

        """
        return self.GetData("performance/riskVolatility")

    def salesFees(self):
        """
        This function retrieves the sales fees of the funds

        Returns:
            dict fees

        Examples:
            >>> Funds("myria", "fr").salesFees()

        """
        if self.asset_type == "etf":
            return {}
        return self.GetData("price/salesFees")

    def sector(self):
        """
        This function retrieves the sector breakdown of the funds, category and index

        Returns:
            dict sector breakdown

        Examples:
            >>> Funds("myria", "fr").sector()

        """
        return self.GetData("portfolio/v2/sector")

    def snapshot(self, currency="EUR"):
        """
        This function returns a snapshot of the fund and asset manager.

        Returns:
            dict snapshot fund and asset manager

        Examples:
            >>> Funds("myria", "fr").snapshot()

        """
        return self.ltData("MFsnapshot", currency=currency)

    def starRatingFundAsc(self):
        """
        This function retrieves the MorningStar rating of the funds
        of the company by ascending order

        Returns:
            dict rating

        Examples:
            >>> Funds("myria", "fr").starRatingFundAsc()

        """

        return self.GetData("parent/mstarRating/StarRatingFundAsc")

    def starRatingFundDesc(self):
        """
        This function retrieves the MorningStar rating of the funds
        of the company by descending order

        Returns:
            dict rating

        Examples:
            >>> Funds("myria", "fr").starRatingFundDesc()

        """

        return self.GetData("parent/mstarRating/StarRatingFundDesc")

    def sustainability(self, currency="EUR"):
        """
        This function retrieves the sustainability data of the fund.

        Returns:
            dict sustanability data

        Examples:
            >>> Funds("myria", "fr").sustainability()

        """
        return self.ltData("sustainability", currency=currency)

    def taxes(self):
        """
        This function retrieves the other fee of the etf

        Returns:
            dict taxes

        Examples:
            >>> Funds("American Century Foc Dynmc Gr ETF").taxes()

        """
        return self.GetData("price/taxes")

    def trailingReturn(self, duration="daily"):
        """
        This function retrieves the trailing return of the funds of the company.

        Args:
        duration (str) : frequency of return can be daily, monthly or quarterly

        Returns:
            dict trailing return

        Raises:
            ValueError whenever the parameter duration is not daily, monthly or quarterly

        Example:
            >>> Funds("myria", "fr").trailingReturn("daily")

        """

        duration_choice = ["daily", "monthly", "quarterly"]
        if duration not in duration_choice:
            raise ValueError(
                f'duration parameter can only take one of the values: {", ".join(duration_choice)}'
            )

        return self.GetData("trailingReturn/v2", {"duration": duration})
